/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.util;

import jalview.bin.Cache;
import jalview.bin.Jalview;
import jalview.gui.IProgressIndicator;
import jalview.gui.LineartOptions;
import jalview.io.JalviewFileChooser;
import jalview.util.dialogrunner.RunResponse;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileOutputStream;
import java.util.concurrent.atomic.AtomicBoolean;

import javax.imageio.ImageIO;
import javax.swing.JOptionPane;

import org.jfree.graphics2d.svg.SVGGraphics2D;
import org.jfree.graphics2d.svg.SVGHints;
import org.jibble.epsgraphics.EpsGraphics2D;

public class ImageMaker
{
  public static final String SVG_DESCRIPTION = "Scalable Vector Graphics";

  public static final String SVG_EXTENSION = "svg";

  public static final String EPS_DESCRIPTION = "Encapsulated Postscript";

  public static final String EPS_EXTENSION = "eps";

  public static final String PNG_EXTENSION = "png";

  public static final String PNG_DESCRIPTION = "Portable  network graphics";

  public static final String HTML_EXTENSION = "html";

  public static final String HTML_DESCRIPTION = "Hypertext Markup Language";

  EpsGraphics2D pg;

  Graphics graphics;

  FileOutputStream out;

  BufferedImage bi;

  TYPE type;

  private IProgressIndicator pIndicator;

  private long pSessionId;

  private boolean headless;

  public enum TYPE
  {
    EPS("EPS", MessageManager.getString("label.eps_file"), EPS_EXTENSION,
            EPS_DESCRIPTION),
    PNG("PNG", MessageManager.getString("label.png_image"), PNG_EXTENSION,
            PNG_DESCRIPTION),
    SVG("SVG", "SVG", SVG_EXTENSION, SVG_DESCRIPTION);

    public final String name;

    public final String label;

    public final String extension;

    public final String description;

    TYPE(String name, String label, String ext, String desc)
    {
      this.name = name;
      this.label = label;
      this.extension = ext;
      this.description = desc;
    }

    public String getName()
    {
      return name;
    }

    public JalviewFileChooser getFileChooser()
    {
      return new JalviewFileChooser(extension, description);
    }

    public String getLabel()
    {
      return label;
    }

  }

  /**
   * Constructor builds the image and writes it to file. If the supplied file
   * name is null, the user is prompted for the output file.
   * 
   * @param parent
   * @param type
   * @param title
   * @param width
   * @param height
   * @param file
   * @param fileTitle
   * @param pIndicator
   * @param pSessionId
   * @param headless
   */
  public ImageMaker(Component parent, TYPE type, String title, int width,
          int height, File file, String fileTitle,
          IProgressIndicator pIndicator, long pSessionId, boolean headless)
  {
    this.pIndicator = pIndicator;
    this.type = type;
    this.pSessionId = pSessionId;
    this.headless = headless;
    if (file == null && !Jalview.isHeadlessMode())
    {
      setProgressMessage(MessageManager.formatMessage(
              "status.waiting_for_user_to_select_output_file", type.name));
      JalviewFileChooser chooser = type.getFileChooser();
      chooser.setFileView(new jalview.io.JalviewFileView());
      chooser.setDialogTitle(title);
      chooser.setToolTipText(MessageManager.getString("action.save"));
      int value = chooser.showSaveDialog(parent);

      if (value == JalviewFileChooser.APPROVE_OPTION)
      {
        Cache.setProperty("LAST_DIRECTORY",
                chooser.getSelectedFile().getParent());
        file = chooser.getSelectedFile();
      }
      else
      {
        setProgressMessage(MessageManager.formatMessage(
                "status.cancelled_image_export_operation", type.name));
      }
    }

    if (file != null)
    {
      try
      {
        out = new FileOutputStream(file);
        setProgressMessage(null);
        setProgressMessage(MessageManager.formatMessage(
                "status.exporting_alignment_as_x_file", type.getName()));
        if (type == TYPE.SVG)
        {
          setupSVG(width, height, fileTitle);
        }
        else if (type == TYPE.EPS)
        {
          setupEPS(width, height, fileTitle);
        }
        else if (type == TYPE.PNG)
        {
          setupPNG(width, height);
        }

      } catch (Exception ex)
      {
        System.out.println("Error creating " + type.getName() + " file.");

        setProgressMessage(MessageManager
                .formatMessage("info.error_creating_file", type.getName()));
      }
    }
  }

  public Graphics getGraphics()
  {
    return graphics;
  }

  /**
   * For SVG or PNG, writes the generated graphics data to the file output
   * stream. For EPS, flushes the output graphics (which is written to file as
   * it is generated).
   */
  public void writeImage()
  {
    try
    {
      switch (type)
      {
      case EPS:
        pg.flush();
        pg.close();
        break;
      case SVG:
        String svgData = ((SVGGraphics2D) getGraphics()).getSVGDocument();
        out.write(svgData.getBytes());
        out.flush();
        out.close();
        break;
      case PNG:
        ImageIO.write(bi, PNG_EXTENSION, out);
        out.flush();
        out.close();
        break;
      }
    } catch (Exception ex)
    {
      ex.printStackTrace();
    }
  }

  /**
   * Generates an EPS image and writes it to the (previously set) output buffer.
   * The user is first prompted for choice of Text or Lineart rendering, unless
   * a preference for this has been set.
   * 
   * @param width
   * @param height
   * @param title
   */
  void setupEPS(int width, int height, String title)
  {
    String renderStyle = Cache.getDefault("EPS_RENDERING",
            "Prompt each time");
    AtomicBoolean textOption = new AtomicBoolean(
            !"Lineart".equals(renderStyle));

    /*
     * configure the action to run on OK in the dialog
     */
    RunResponse okAction = new RunResponse(JOptionPane.OK_OPTION)
    {
      @Override
      public void run()
      {
        writeEPS(width, height, title, textOption.get());
      }
    };

    /*
     * Prompt for character rendering style if preference is not set
     */
    if (renderStyle.equalsIgnoreCase("Prompt each time")
            && !(System.getProperty("java.awt.headless") != null && System
                    .getProperty("java.awt.headless").equals("true")))
    {
      LineartOptions epsOption = new LineartOptions("EPS_RENDERING",
              TYPE.EPS.getName(), textOption);
      epsOption.setResponseAction(new RunResponse(JOptionPane.NO_OPTION)
      {
        @Override
        public void run()
        {
          setProgressMessage(MessageManager.formatMessage(
                  "status.cancelled_image_export_operation", "EPS"));
        }
      });
      epsOption.setResponseAction(okAction);
      epsOption.showDialog();
      /* no code here - JalviewJS cannot execute it */
    }
    else
    {
      /*
       * else (if preference set) just do the export action
       */
      writeEPS(width, height, title, textOption.get());
    }
  }

  /**
   * Sets up a graphics object for the PNG image to be written on
   * 
   * @param width
   * @param height
   */
  void setupPNG(int width, int height)
  {
    bi = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
    graphics = bi.getGraphics();
    Graphics2D ig2 = (Graphics2D) graphics;
    ig2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
            RenderingHints.VALUE_ANTIALIAS_ON);
    setProgressMessage(MessageManager
            .formatMessage("status.export_complete", type.getName()));
  }

  /**
   * Sets up a graphics object for the SVG image to be written on. The user is
   * first prompted for choice of Text or Lineart rendering, unless a preference
   * for this has been set.
   * 
   * @param width
   * @param height
   * @param title
   */
  void setupSVG(int width, int height, String title)
  {
    String renderStyle = Cache.getDefault("SVG_RENDERING",
            "Prompt each time");
    AtomicBoolean textOption = new AtomicBoolean(
            !"Lineart".equals(renderStyle));

    /*
     * configure the action to run on OK in the dialog
     */
    RunResponse okAction = new RunResponse(JOptionPane.OK_OPTION)
    {
      @Override
      public void run()
      {
        setupSVG(width, height, textOption.get());
      }
    };

    /*
     * Prompt for character rendering style if preference is not set
     */
    if (renderStyle.equalsIgnoreCase("Prompt each time")
            && !(System.getProperty("java.awt.headless") != null && System
                    .getProperty("java.awt.headless").equals("true")))
    {
      LineartOptions svgOption = new LineartOptions("SVG_RENDERING",
              TYPE.SVG.getName(), textOption);
      svgOption.setResponseAction(new RunResponse(JOptionPane.NO_OPTION)
      {
        @Override
        public void run()
        {
          setProgressMessage(MessageManager.formatMessage(
                  "status.cancelled_image_export_operation", "SVG"));
        }
      });
      svgOption.setResponseAction(okAction);
      svgOption.showDialog();
      /* no code here - JalviewJS cannot execute it */
    }
    else
    {
      /*
       * else (if preference set) just do the export action
       */
      setupSVG(width, height, textOption.get());
    }
  }

  void setProgressMessage(String message)
  {
    if (pIndicator != null && !headless)
    {
      pIndicator.setProgressBar(message, pSessionId);
    }
  }

  /**
   * A helper method to configure the SVG output graphics, with choice of Text
   * or Lineart character rendering
   * 
   * @param width
   * @param height
   * @param textOption
   *          true for Text, false for Lineart
   */
  protected void setupSVG(int width, int height, boolean textOption)
  {
    SVGGraphics2D g2 = new SVGGraphics2D(width, height);
    if (!textOption) // Lineart selected
    {
      g2.setRenderingHint(SVGHints.KEY_DRAW_STRING_TYPE,
              SVGHints.VALUE_DRAW_STRING_TYPE_VECTOR);
    }
    setProgressMessage(MessageManager
            .formatMessage("status.export_complete", type.getName()));
    graphics = g2;
  }

  /**
   * A helper method that sets up the EPS graphics output with user choice of
   * Text or Lineart character rendering
   * 
   * @param width
   * @param height
   * @param title
   * @param textOption
   *          true for Text, false for Lineart
   */
  protected void writeEPS(int width, int height, String title,
          boolean textOption)
  {
    try
    {
      pg = new EpsGraphics2D(title, out, 0, 0, width, height);
      Graphics2D ig2 = pg;
      ig2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
              RenderingHints.VALUE_ANTIALIAS_ON);
      pg.setAccurateTextMode(!textOption); // true = Lineart
      graphics = pg;
      setProgressMessage(MessageManager
              .formatMessage("status.export_complete", type.getName()));
    } catch (Exception ex)
    {
      System.err.println("Error writing PNG: " + ex.toString());
    }
  }

  static JalviewFileChooser getSVGChooser()
  {
    if (Jalview.isHeadlessMode())
    {
      return null;
    }
    return new JalviewFileChooser(SVG_EXTENSION, SVG_DESCRIPTION);
  }
}
